import logging, time

from .Upload import *
from .Variable import *
from .api_request import make_request, make_paginated_request


class Table:
    def __init__(self, identifier, properties=None):
        self.uri = f"/tables/{identifier}"
        self.properties = properties

    def __getitem__(self, key):
        return self.properties[key]

    def __str__(self):
        return json.dumps(self.properties, indent=2)

    def list_uploads(self):
        uploads = make_paginated_request(path=f"{self.uri}/uploads")
        # TODO: for each upload, Upload(res)
        return

    def list_variables(self):
        variables = make_paginated_request(path=f"{self.uri}/variables")
        # TODO: for each variable, Variable(res)
        return

    def update(self, *, name, description, merge_strategy):
        response = make_request(
            method="PATCH",
            path=f"{self.uri}",
            payload={
                "name": name,
                "description": description,
                "mergeStrategy": merge_strategy,
            },
        )
        self.properties = response
        return

    def create_upload(self, *, name, data, type, remove_on_fail=True):
        response = make_request(
            method="POST",
            path=f"{self.uri}/uploads",
            payload={"name": name, "type": type},
        )
        upload = Upload(uri=response["uri"])
        try:
            upload.upload_file(data)
            while True:
                time.sleep(2)
                upload.get()
                if upload["status"] == "completed":
                    break
                elif upload["status"] == "failed":
                    raise Exception(upload["errorMessage"])
                else:
                    logging.debug("Upload is still in progress...")
        except Exception as e:
            if remove_on_fail:
                print("An error occurred. Deleting upload.")
                upload.delete()
            raise e

        return upload

    def download(self):
        # TODO
        return

    def delete(self):
        make_request(
            method="DELETE",
            path=self.uri,
        )
        return



import requests
import logging
import os
import re
import json

api_endpoint = (
    "https://redivis.com/api/v1"
    if os.getenv("REDIVIS_API_ENDPOINT") is None
    else os.getenv("REDIVIS_API_ENDPOINT")
)

verifySSL = False if api_endpoint.find("https://localhost", 0) == 0 else True


def make_request(*, method, path, query=None, payload=None, parse_payload=True):
    method = method.lower()
    url = f"{api_endpoint}{path}"

    headers = {"Authorization": f"Bearer {__get_access_token()}"}

    logging.debug(f"Making API '{method}' request to '{url}'")

    if parse_payload and payload:
        payload = json.dumps(payload)

    r = getattr(requests, method)(
        url, headers=headers, params=query, verify=verifySSL, data=payload
    )

    if r.text != "OK":
        json_response = r.json()
        if r.status_code >= 400:
            raise requests.HTTPError(json_response)
        return json_response

    else:
        r.raise_for_status()
        return

    # return __invert_case(json_response, __camel_to_snake)


def make_paginated_request(*, path, query=None, max_results=100):
    # TODO: proper pagination with max_results. How to use iterators?
    headers = {"Authorization": f"Bearer {__get_access_token()}"}
    url = f"{api_endpoint}{path}"
    logging.debug(f"Making paginated API request to '{url}'")
    requests.get(url, headers=headers, params={"maxResults": max_results}.update(query))


def __get_access_token():
    if os.getenv("REDIVIS_API_TOKEN") is None:
        raise EnvironmentError(
            "The environment variable REDIVIS_API_TOKEN must be set."
        )
    return os.environ["REDIVIS_API_TOKEN"]


def __camel_case_to_snake_case(json_response):

    return json_response


def __camel_to_snake(name):
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def __snake_to_camel(name):
    return re.sub("_([a-z])", lambda match: match.group(1).upper(), name)


def __invert_case(dict_obj, inverter):
    assert type(dict_obj) == dict
    converted_dict_obj = {}
    for name in dict_obj:
        new_name = inverter(name)
        value = dict_obj[name]

        if type(value) == dict:
            converted_dict_obj[new_name] = __invert_case(value, inverter)
        else:
            converted_dict_obj[new_name] = dict_obj[name]
    return converted_dict_obj



try:
    import pkg_resources

    pkg_resources.declare_namespace(__name__)
except ImportError:
    import pkgutil

    __path__ = pkgutil.extend_path(__path__, __name__)

from .common.Dataset import *
from .common.Table import *




from google.cloud import bigquery
from google.auth import credentials
import os

class Credentials(credentials.AnonymousCredentials):
	def __init__(self, token):
		self.token = token
	def apply(self, headers):
		headers['Authorization'] = 'Bearer {}'.format(self.token)
	def before_request(self, request, method, url, headers):
		self.apply(headers)


class Client(bigquery.Client):
	def __init__(
		self,
		api_endpoint = "https://redivis.com/api/v1"
	):
		credential = Credentials(os.getenv('REDIVIS_API_TOKEN'))
		super(Client, self).__init__(
			credentials=credential, project="_", client_options={"api_endpoint": api_endpoint}
		)



import json
import requests

from .Table import *
from .api_request import make_request


class Dataset:
    def __init__(self, identifier, properties=None):
        self.uri = f"/datasets/{identifier}"
        self.properties = properties

    def __getitem__(self, key):
        return self.properties[key]

    def __str__(self):
        return json.dumps(self.properties, indent=2)

    def exists(self):
        try:
            make_request(method="GET", path=self.uri)
        except requests.HTTPError as err:
            # TODO: check status code
            return False

        return True

    def get(self):
        self.properties = make_request(method="GET", path=self.uri)

        self.uri = self.properties["uri"]
        return self

    def update(self, name, public_access_level, description):
        self.properties = make_request(
            method="POST",
            path=self.uri,
            payload={
                "name": name,
                "publicAccessLevel": public_access_level,
                "description": description,
            },
        )
        return self

    def list_tables(self, max_results):
        # TODO
        return

    def list_versions(self, max_results):
        # TODO
        return

    def create_version(self, ignore_if_exists=False):
        if not self.properties or not hasattr(self.properties, "nextVersion"):
            self.get()

        if not self.properties["nextVersion"]:
            response = make_request(method="POST", path=f"{self.uri}/versions")
            return Dataset(response["datasetUri"].replace("/datasets/", "")).get()
        elif ignore_if_exists:
            return Dataset(
                self.properties["nextVersion"]["datasetUri"].replace("/datasets/", "")
            ).get()
        else:
            raise Exception(
                f"Next version already exists at {self.properties['nextVersion']['datasetUri']}. To avoid this error, set argumeent ignore_if_exists to True"
            )

    def release(self):
        response = make_request(
            method="POST",
            path=f"{self.uri}/versions/next/release",
        )
        self.uri = response["datasetUri"]
        return self.get()

    def create_table(self, *, name, description, upload_merge_strategy):
        response = make_request(
            method="POST",
            path=f"{self.uri}/tables",
            payload={
                "name": name,
                "description": description,
                "uploadMergeStrategy": upload_merge_strategy,
            },
        )
        return Table(response["uri"].replace("/tables/", ""), response)

    def delete(self):
        make_request(
            method="DELETE",
            path=self.uri,
        )
        return


def create_dataset(
    *, user=None, organization=None, name, public_access_level, description=None
):
    if user:
        path = f"/users/{user}/datasets"
    elif organization:
        path = f"/organizations/{organization}/datasets"
    else:
        raise Exception("Must provide a user or organization")

    response = make_request(
        method="POST",
        path=path,
        payload={
            "name": name,
            "publicAccessLevel": public_access_level,
            "description": description,
        },
    )

    return Dataset(
        identifier=response["uri"].replace("/datasets/", ""), properties=response
    )
